"""Core data models for LLMling agent."""

from __future__ import annotations

from llmling_agent.models.agents import AgentConfig
from llmling_agent.models.manifest import AgentsManifest


__all__ = ["AgentConfig", "AgentsManifest"]
